import datetime

from decimal import Decimal

from sqlalchemy.inspection import inspect

from .config import ALLOWED_METHODS
from .config import MODEL_VERSION


class Model(object):
    __url__ = None
    __table__ = None
    __version__ = MODEL_VERSION
    __methods__ = ALLOWED_METHODS

    def __str__(self):
        """

        :return:
        """
        return str(getattr(self, self.primary_key()))

    @classmethod
    def required(cls):
        """

        :return:
        """
        columns = []
        for c in cls.__table__.columns:
            is_autoincrement = 'int' in str(c.type).lower() and c.autoincrement
            if (not c.nullable and not c.primary_key) or (c.primary_key and not is_autoincrement):
                columns.append(c.name)
        return columns

    @classmethod
    def optional(cls):
        """

        :return:
        """
        columns = []
        for c in cls.__table__.columns:
            if c.nullable:
                columns.append(c.name)
        return columns

    @classmethod
    def primary_key(cls):
        """

        :return:
        """
        return list(cls.__table__.primary_key.columns)[0].key

    def to_dict(self):
        """

        :return:
        """
        result_dict = {}
        for c in self.__table__.columns.keys():
            value = result_dict[c] = getattr(self, c, None)
            if isinstance(value, Decimal):
                result_dict[c] = float(result_dict[c])
            elif isinstance(value, datetime.datetime):
                result_dict[c] = value.isoformat()
        return result_dict

    def links(self):
        """

        :return:
        """
        link_dict = {'self': self.resource_uri()}
        for r in inspect(self.__class__).relationships:
            if 'collection' not in r.key:
                instance = getattr(self, r.key)
                if instance:
                    link_dict[str(r.key)] = instance.resource_uri()
        return link_dict

    def resource_uri(self):
        """

        :return:
        """
        return self.__url__ + '/' + str(getattr(self, self.primary_key()))

    def update(self, attributes):
        """

        :param attributes:
        :return:
        """
        for attribute in attributes:
            setattr(self, attribute, attributes[attribute])
        return self

    @classmethod
    def description(cls):
        """

        :return:
        """
        description = {}
        for c in cls.__table__.columns:
            column_description = str(c.type)
            if not c.nullable:
                column_description += ' (required)'
            description[c.name] = column_description
        return description



from flask import abort
from flask import request

from sqlalchemy import asc as ASC
from sqlalchemy import desc as DESC

from .wrapper import resp_json

from .config import GRAMMAR
from .config import ARGUMENT
from .config import HTTP_STATUS


def validate_entity(model, data):
    """

    :param model:
    :param data:
    """
    fields = model.required() + model.optional()
    unknown = [k for k in data if k not in fields]
    missing = set(model.required()) - set(data)

    if len(unknown) or len(missing):
        abort(
            resp_json({
                'unknown': unknown,
                'missing': list(missing)
            }, code=HTTP_STATUS.UNPROCESSABLE_ENTITY)
        )


def parsing_query_string(model):
    """

    :param model:
    :return:
    """
    order = []
    fields = []
    filters = []
    invalid = []

    for k, v in request.args.items():
        if k in ARGUMENT.STATIC.__dict__.keys():
            continue
        if hasattr(model, k):
            attribute = getattr(model, k)

            if v.startswith(GRAMMAR.NOT_LIKE):
                filters.append(attribute.notilike(v[2:], escape='/'))
            elif v.startswith(GRAMMAR.LIKE):
                filters.append(attribute.ilike(v[1:], escape='/'))
            else:
                items = v.split(GRAMMAR.SEP)
                if len(items) > 1:
                    if items[0].startswith(GRAMMAR.NOT):
                        items[0] = items[0].lstrip(GRAMMAR.NOT)
                        in_statement = ~attribute.in_(items)
                    else:
                        in_statement = attribute.in_(items)
                    filters.append(in_statement)
                else:
                    if v.startswith(GRAMMAR.NOT):
                        filters.append(attribute != (None if v == GRAMMAR.NOT_NULL else v.lstrip(GRAMMAR.NOT)))
                    else:
                        filters.append(attribute == (None if v == GRAMMAR.NULL else v.lstrip('\\')))

        elif k == ARGUMENT.DYNAMIC.sort:
            for item in v.split(GRAMMAR.NOT):
                direction = DESC if item.startswith(GRAMMAR.REVERSE) else ASC
                item = item.lstrip(GRAMMAR.REVERSE)

                if not hasattr(model, item):
                    invalid.append(item)
                else:
                    order.append(direction(getattr(model, item)))

        elif k == ARGUMENT.DYNAMIC.fields:
            fields = v.split(GRAMMAR.SEP)
            for item in fields:
                if not hasattr(model, item):
                    invalid.append(item)
        else:
            invalid.append(k)

    if len(invalid) > 0:
        abort(resp_json({'invalid': invalid}, code=HTTP_STATUS.BAD_REQUEST))

    return fields, model.query.filter(*filters).order_by(*order)



from flask import Blueprint

from flask_json import as_json

from sqlalchemy.sql import sqltypes
from sqlalchemy.ext.automap import automap_base
from sqlalchemy.ext.declarative import declarative_base

from flask_autocrud.model import Model
from flask_autocrud.service import Service


class AutoCrud(object):
    def __init__(self, app=None, db=None, admin=None, view=None,
                 exclude_tables=None, user_models=None, schema=None):
        """

        :param db:
        :param app:
        :param admin:
        :param exclude_tables:
        :param user_models:
        :param schema:
        """
        self._db = db
        self._admin = admin
        self._view = view
        self._api = None
        self._automap_model = None
        self._exclude_tables = exclude_tables
        self._user_models = user_models
        self._schema = schema

        if app is not None:
            self._app = app
            self.init_app(self._app, self._db, self._admin, self._view)
        else:
            self._app = None

    def init_app(self, app, db, admin=None, view=None,
                 exclude_tables=None, user_models=None, schema=None):
        """

        :param app:
        :param db:
        :param admin:
        :param view:
        :param exclude_tables:
        :param user_models:
        :param schema:
        :return:
        """
        self._app = app
        self._db = db
        self._admin = admin
        self._view = view
        self._exclude_tables = exclude_tables
        self._user_models = user_models
        self._schema = schema

        if self._db is None:
            raise AttributeError("""
                You can not create AutoCrud without an SQLAlchemy instance.
                Please consider to use the init_app method instead
            """)

        if self._admin and not self._view:
            raise AttributeError("""
                You can not create AutoCrud with Admin but without ModelView instance.
                admin and view arguments are required together
            """)

        self._app.classes = []
        self._automap_model = automap_base(declarative_base(cls=(db.Model, Model)))
        self._config()

        with self._app.app_context():
            if self._user_models:
                if any([issubclass(cls, self._automap_model) for cls in self._user_models]):
                    self._automap_model.prepare(self._db.engine, reflect=True, schema=self._schema)

                for user_model in self._user_models:
                    self._register_model(user_model)
            else:
                self._automap_model.prepare(self._db.engine, reflect=True, schema=self._schema)

                for cls in self._automap_model.classes:
                    if self._exclude_tables and cls.__table__.name in self._exclude_tables:
                        continue

                    if self._app.config['AUTOCRUD_READ_ONLY']:
                        cls.__methods__ = {'GET'}

                    self._register_model(cls)

        @self._app.route(self._app.config['AUTOCRUD_BASE_URL'])
        @as_json
        def index():
            """

            :return:
            """
            routes = {}
            for cls in self._app.classes:
                routes[cls.__model__.__name__] = '{}{{/{}}}'.format(
                    cls.__model__.__url__,
                    cls.__model__.primary_key()
                )
            return routes

        self._app.register_blueprint(self._api)

    def _register_model(self, cls):
        """

        :param cls:
        """
        if self._admin is not None:
            self._admin.add_view(self._view(cls, self._db.session))

        cls.__url__ = '{}{}'.format(
            self._app.config['AUTOCRUD_BASE_URL'],
            cls.__name__.lower()
        )
        cols = list(cls().__table__.primary_key.columns)

        service_class = type(
            cls.__name__ + 'Service',
            (Service,),
            {
                '__model__': cls,
                '__db__': self._db,
                '__collection_name__': cls.__name__
            }
        )

        primary_key_type = 'string'
        if len(cols) == 1:
            col_type = cols[0].type
            if isinstance(col_type, sqltypes.String):
                primary_key_type = 'string'
            elif isinstance(col_type, sqltypes.Integer):
                primary_key_type = 'int'
            elif isinstance(col_type, sqltypes.Numeric):
                primary_key_type = 'float'

        cls = service_class
        endpoint = self._api
        self._app.classes.append(cls)

        view_func = cls.as_view(service_class.__name__.lower())
        methods = set(cls.__model__.__methods__)

        if 'GET' in methods:
            endpoint.add_url_rule(
                cls.__model__.__url__ + '/',
                defaults={'resource_id': None},
                view_func=view_func,
                methods=['GET']
            )
            if self._app.config['AUTOCRUD_METADATA_ENABLED'] is True:
                endpoint.add_url_rule(
                    '{resource}/meta'.format(resource=cls.__model__.__url__),
                    view_func=view_func,
                    methods=['GET']
                )

        if 'POST' in methods:
            endpoint.add_url_rule(
                cls.__model__.__url__ + '/',
                view_func=view_func,
                methods=['POST'],
                strict_slashes=False
            )

        endpoint.add_url_rule(
            '{resource}/<{pk_type}:{pk}>'.format(
                resource=cls.__model__.__url__,
                pk='resource_id',
                pk_type=primary_key_type
            ),
            view_func=view_func,
            methods=methods - {'POST'}
        )

    def _config(self):
        """

        """
        self._app.config.setdefault('AUTOCRUD_METADATA_ENABLED', True)
        self._app.config.setdefault('AUTOCRUD_READ_ONLY', False)
        self._app.config.setdefault('AUTOCRUD_BASE_URL', '/')
        self._app.config.setdefault('AUTOCRUD_SUBDOMAIN', None)

        self._app.config.setdefault('JSON_ADD_STATUS', False)
        self._app.config.setdefault('JSON_DECODE_ERROR_MESSAGE', 'Invalid JSON')

        self._api = Blueprint(
            'flask_autocrud',
            __name__,
            subdomain=self._app.config['AUTOCRUD_SUBDOMAIN']
        )



from flask import abort
from flask import request

from flask.views import MethodView

from .wrapper import get_json
from .wrapper import resp_csv
from .wrapper import resp_json
from .wrapper import no_content
from .wrapper import response_with_links
from .wrapper import response_with_location

from .validators import validate_entity
from .validators import parsing_query_string


from .config import ARGUMENT
from .config import HTTP_STATUS


class Service(MethodView):
    __db__ = None
    __model__ = None
    __collection_name__ = 'resources'

    def delete(self, resource_id):
        """

        :param resource_id:
        :return:
        """
        model = self.__model__
        session = self.__db__.session()

        resource = model.query.get(resource_id)
        if not resource:
            abort(resp_json({'message': 'Not Found'}, code=HTTP_STATUS.NOT_FOUND))

        session.delete(resource)
        session.commit()
        return no_content()

    def get(self, resource_id=None):
        """

        :param resource_id:
        :return:
        """
        response = []
        model = self.__model__
        page = request.args.get(ARGUMENT.STATIC.page)
        limit = request.args.get(ARGUMENT.STATIC.limit)
        export = request.args.get(ARGUMENT.STATIC.export)

        if resource_id is not None:
            resource = model.query.get(resource_id)
            if not resource:
                abort(resp_json({'message': 'Not Found'}, code=HTTP_STATUS.NOT_FOUND))
            return response_with_links(resource)

        if request.path.endswith('meta'):
            return resp_json(model.description())

        fields, statement = parsing_query_string(model)

        if page is not None:
            resources = statement.paginate(page=page, per_page=limit).items
        else:
            resources = statement.limit(limit).all()

        for r in resources:
            item = r.to_dict()
            item_keys = item.keys()
            if fields:
                for k in set(item_keys) - set(fields):
                    item.pop(k)
            response.append(item)

        response_builder = resp_csv if export else resp_json
        return response_builder(response, self.__collection_name__)

    def patch(self, resource_id):
        """

        :param resource_id:
        :return:
        """
        model = self.__model__
        session = self.__db__.session()

        data = get_json()
        validate_entity(model, data)

        resource = model.query.get(resource_id)
        if not resource:
            abort(resp_json({'message': 'Not Found'}, code=HTTP_STATUS.NOT_FOUND))

        resource.update(data)
        session.merge(resource)
        session.commit()

        return response_with_links(resource)

    def post(self):
        """

        :return:
        """
        model = self.__model__
        session = self.__db__.session()

        data = get_json()
        validate_entity(model, data)

        resource = model.query.filter_by(**data).first()
        if not resource:
            resource = model(**data)
            session.add(resource)
            session.commit()
            code = HTTP_STATUS.CREATED
        else:
            code = HTTP_STATUS.CONFLICT

        return response_with_location(resource, code)

    def put(self, resource_id):
        """

        :param resource_id:
        :return:
        """
        model = self.__model__
        session = self.__db__.session()

        data = get_json()
        validate_entity(model, data)

        resource = model.query.get(resource_id)
        if resource:
            resource.update(data)
            session.merge(resource)
            session.commit()

            return response_with_links(resource)

        resource = model(**data)
        session.add(resource)
        session.commit()

        return response_with_links(resource, HTTP_STATUS.CREATED)



"""
Flask-AutoCRUD
-------------

Automatically generate a RESTful API service for CRUD operation on database
"""

from setuptools import setup


with open("README.rst", "r") as fh:
    long_description = fh.read()

setup(
    name='Flask-AutoCRUD',
    version='1.1.2',
    url='https://github.com/cs91chris/flask_autocrud/',
    license='MIT',
    author='cs91chris',
    author_email='cs91chris@voidbrain.me',
    description='Automatically generate a RESTful API service for CRUD operation on database',
    long_description=long_description,
    packages=['flask_autocrud'],
    zip_safe=False,
    include_package_data=True,
    platforms='any',
    install_requires=[
        'Flask==1.0.2',
        'Flask-Admin==1.5.3',
        'Flask-SQLAlchemy==2.3.2',
        'Flask-JSON==0.3.3'
    ],
    classifiers=[
        'Environment :: Web Environment',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License',
        'Operating System :: OS Independent',
        'Programming Language :: Python :: 3',
        'Topic :: Internet :: WWW/HTTP :: Dynamic Content',
        'Topic :: Software Development :: Libraries :: Python Modules'
    ]
)



class GRAMMAR:
    SEP = ';'
    NOT = '!'
    REVERSE = '-'
    NULL = 'null'
    LIKE = '%'
    NOT_NULL = NOT + NULL
    NOT_LIKE = NOT + LIKE


class ARGUMENT:
    class STATIC:
        page = 'page'
        limit = 'limit'
        export = 'export'

    class DYNAMIC:
        sort = 'sort'
        fields = 'fields'


class HTTP_STATUS:
    SUCCESS = 200
    CREATED = 201
    NO_CONTENT = 204
    BAD_REQUEST = 400
    NOT_FOUND = 404
    CONFLICT = 409
    UNPROCESSABLE_ENTITY = 422
    INTERNAL_SERVER_ERROR = 500


ALLOWED_METHODS = {
    'GET',
    'POST',
    'PUT',
    'PATCH',
    'DELETE',
    'HEAD',
    'OPTIONS'
}

MODEL_VERSION = '1'
